from context import classes, errors
from hashlib import sha256, sha3_256
from random import randint
import unittest


class TestMerkle(unittest.TestCase):
    """Test suite for the Merkle project."""
    def setUp(self) -> None:
        self.original_hash_function = classes.get_hash_function()

    def tearDown(self) -> None:
        classes.set_hash_function(self.original_hash_function)

    def test_merkle_has_class_Tree(self):
        assert hasattr(classes, 'Tree')
        assert isinstance(classes.Tree, type)

    def test_Tree_joins_left_and_right_into_root(self):
        classes.set_hash_function(lambda data: sha3_256(data).digest())
        left = sha3_256(b'hello').digest()
        right = sha3_256(b'world').digest()
        joined = sha3_256(left + right).digest()
        tree = classes.Tree(left, right)

        assert hasattr(tree, 'root')
        assert type(tree.root) is bytes
        assert tree.root == joined

    def test_Tree_from_leaves_hashes_and_joins_leaves(self):
        leaves = [b'hello', b'world']
        left_hash = sha256(leaves[0]).digest()
        right_hash = sha256(leaves[1]).digest()
        root = sha256(left_hash + right_hash).digest()

        assert hasattr(classes.Tree, 'from_leaves')
        tree = classes.Tree.from_leaves(leaves)
        assert tree.left == left_hash
        assert tree.right == right_hash
        assert tree.root == root

    def test_Tree_from_leaves_tressas_at_least_two_leaves(self):
        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.from_leaves([b'123'])
        assert str(e.exception) == 'must have at least 2 leaves'

    def test_Tree_from_leaves_joins_any_number_of_leaves(self):
        roots = set()

        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = classes.Tree.from_leaves(leaves)
            assert tree.root not in roots
            roots.add(tree.root)

    def test_Tree_instance_serializes_to_dict(self):
        tree = classes.Tree(b'left', b'right')
        assert hasattr(tree, 'to_dict') and callable(tree.to_dict)
        serialized = tree.to_dict()
        assert type(serialized) is dict

    def test_Tree_from_dict_deserializes_and_returns_instance(self):
        tree = classes.Tree(b'left', b'right')
        serialized = tree.to_dict()
        assert hasattr(classes.Tree, 'from_dict')
        deserialized = classes.Tree.from_dict(serialized)
        assert type(deserialized) is classes.Tree
        assert tree == deserialized

    def test_Tree_from_dict_raises_errors_for_invalid_params(self):
        tree = classes.Tree(b'left', b'right')
        serialized = tree.to_dict()

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.from_dict('not a dict')
        assert str(e.exception) == 'data must be dict type'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.from_dict({})
        assert str(e.exception) == 'data must have one key'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.from_dict({**serialized, 'what': 'huh'})
        assert str(e.exception) == 'data must have one key'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.from_dict({"3213": [1,2,3]})
        assert str(e.exception) == 'data[root] must have left and right branch'

        with self.assertRaises(errors.SecurityError) as e:
            key = list(serialized.keys())[0]
            classes.Tree.from_dict({"2323": serialized[key]})
        assert str(e.exception) == 'root mismatch'

    def test_Tree_instance_serializes_to_json(self):
        tree = classes.Tree(b'left', b'right')
        assert hasattr(tree, 'to_json') and callable(tree.to_json)
        serialized = tree.to_json()
        assert type(serialized) is str

    def test_Tree_from_json_unserializes_and_returns_instance(self):
        tree = classes.Tree(b'left', b'right')
        serialized = tree.to_json()
        assert hasattr(classes.Tree, 'from_json')
        deserialized = classes.Tree.from_json(serialized)
        assert type(deserialized) is classes.Tree
        assert tree == deserialized

    def test_Tree_prove_produces_list_of_bytes_proof(self):
        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = classes.Tree.from_leaves(leaves)
            proof = tree.prove(randint(0, i-1).to_bytes(2, 'big'))
            proof_verbose = tree.prove(randint(0, i-1).to_bytes(2, 'big'), verbose=True)
            assert type(proof) is list
            assert type(proof_verbose) is list
            for step in proof:
                assert type(step) is bytes
            for step in proof_verbose:
                assert type(step) is bytes

    def test_Tree_prove_raises_errors_for_invalid_params(self):
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = classes.Tree.from_leaves(leaves)

        with self.assertRaises(errors.UsagePreconditionError) as e:
            tree.prove('not bytes')
        assert str(e.exception) == 'leaf must be bytes'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            tree.prove(b'not in tree')
        assert str(e.exception) == 'the given leaf was not found in the tree'

    def test_verbose_proof_is_longer_and_has_2_load_steps_after_hash(self):
        leaves = [n.to_bytes(2, 'big') for n in range(17)]
        tree = classes.Tree.from_leaves(leaves)
        leaf_to_prove = randint(3, 15).to_bytes(2, 'big')
        proof = tree.prove(leaf_to_prove)
        verbose = tree.prove(leaf_to_prove, verbose=True)

        assert len(verbose) > len(proof)

        # normal proof
        assert classes.ProofOp.load_left.value in (proof[0][:1], proof[1][:1])
        assert classes.ProofOp.load_right.value in (proof[0][:1], proof[1][:1])
        assert proof[2][:1] in (
            classes.ProofOp.hash_left.value,
            classes.ProofOp.hash_right.value
        )
        if proof[2][:1] == classes.ProofOp.hash_left.value:
            assert proof[3][:1] != classes.ProofOp.load_left.value
        else:
            assert proof[3][:1] != classes.ProofOp.load_right.value

        # verbose proof
        assert classes.ProofOp.load_left.value in (verbose[0][:1], verbose[1][:1])
        assert classes.ProofOp.load_right.value in (verbose[0][:1], verbose[1][:1])
        assert verbose[2][:1] in (
            classes.ProofOp.hash_left.value,
            classes.ProofOp.hash_right.value
        )
        if verbose[2][:1] == classes.ProofOp.hash_left.value:
            assert verbose[3][:1] == classes.ProofOp.load_left.value
        else:
            assert verbose[3][:1] == classes.ProofOp.load_right.value

    def test_Tree_verify_executes_without_error_for_valid_proof(self):
        for i in range(2, 300):
            leaves = [n.to_bytes(2, 'big') for n in range(i)]
            tree = classes.Tree.from_leaves(leaves)
            leaf = randint(0, i-1).to_bytes(2, 'big')
            proof = tree.prove(leaf)
            classes.Tree.verify(tree.root, leaf, proof)

    def test_Tree_verify_raises_errors_for_invalid_params(self):
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = classes.Tree.from_leaves(leaves)
        leaf = leaves[3]
        proof = tree.prove(leaf)

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.verify('tree.root', leaf, proof)
        assert str(e.exception) == 'root must be bytes'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.verify(tree.root, 'leaf', proof)
        assert str(e.exception) == 'leaf must be bytes'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            classes.Tree.verify(tree.root, leaf, {'not': 'list'})
        assert str(e.exception) == 'proof must be list of bytes'

        with self.assertRaises(errors.UsagePreconditionError) as e:
            wrong_proof = ['not bytes']
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == 'proof must be list of bytes'

    def test_Tree_verify_raises_errors_for_invalid_proofs(self):
        leaves = [n.to_bytes(2, 'big') for n in range(13)]
        tree = classes.Tree.from_leaves(leaves)
        leaf = leaves[3]
        proof = tree.prove(leaf)

        with self.assertRaises(errors.SecurityError) as e:
            classes.Tree.verify(tree.root, leaf + b'1', proof)
        assert str(e.exception) == 'proof does not reference leaf'

        with self.assertRaises(errors.SecurityError) as e:
            wrong_proof = proof[1:]
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == 'proof does not reference leaf'

        with self.assertRaises(errors.SecurityError) as e:
            wrong_proof = proof[:-1]
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == 'proof missing final_hash op'

        with self.assertRaises(errors.SecurityError) as e:
            wrong_proof = [*proof]
            wrong_proof[-1] = wrong_proof[-1] + b'1'
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == 'proof does not reference root'

        with self.assertRaises(ValueError) as e:
            wrong_proof = [*proof]
            wrong_proof[1] = b'\x99' + wrong_proof[1]
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == "b'\\x99' is not a valid ProofOp"

        with self.assertRaises(errors.SecurityError) as e:
            wrong_proof = [*proof]
            wrong_proof[1] = wrong_proof[1] + b'\x99'
            classes.Tree.verify(tree.root, leaf, wrong_proof)
        assert str(e.exception) == "final hash does not match"

    def test_Tree_verify_does_not_validate_malicious_proof(self):
        leaves = [b'leaf0', b'leaf1', b'leaf2']
        tree = classes.Tree.from_leaves(leaves)
        legit_proof = tree.prove(b'leaf0')

        # first instruction in legit_proof is a load_left operation
        assert legit_proof[0][:1] == classes.ProofOp.load_left.value

        # try to trick the validator by inserting malicious leaf then overwriting
        # with the load_left instruction from the legit_proof, then continuing
        # with the legit_proof
        malicious_proof = [
            classes.ProofOp.load_left.value + sha256(b'malicious leaf').digest(),
            *legit_proof
        ]

        with self.assertRaises(errors.SecurityError) as e:
            # raises errors.SecurityError to prevent proof hijacking
            classes.Tree.verify(tree.root, b'malicious leaf', malicious_proof)
        assert str(e.exception) == 'generated hash does not match next step in proof'

        # try to trick the validator by using a proof for a different tree
        malicious_proof = classes.Tree.from_leaves([b'malicious', b'leaves']).prove(b'malicious')

        with self.assertRaises(errors.SecurityError) as e:
            classes.Tree.verify(tree.root, b'malicious', malicious_proof)
        assert str(e.exception) == 'proof does not reference root'

    def test_e2e_arbitrary_branching(self):
        leaves = [sha256(n.to_bytes(2, 'big')).digest() for n in range(13)]

        tree = classes.Tree(leaves[0], leaves[1])
        for i in range(2, len(leaves)):
            if randint(0, 1) == 0:
                tree = classes.Tree(tree, leaves[i])
            else:
                tree = classes.Tree(leaves[i], tree)

        leaf = leaves[randint(0, len(leaves)-1)]
        proof1 = tree.prove(leaf)
        proof2 = tree.prove(leaf, verbose=True)
        classes.Tree.verify(tree.root, leaf, proof1)
        classes.Tree.verify(tree.root, leaf, proof2)


if __name__ == '__main__':
    unittest.main()



class ImplementationError(BaseException):
    ...

class UsagePreconditionError(BaseException):
    ...

class SecurityError(BaseException):
    ...


def tressa(condition: bool, error_message: str) -> None:
    """Raises a UsagePreconditionError with the given error_message if
        the condition is False. Replacement for assert statements and
        AssertionError.
    """
    if not condition:
        raise UsagePreconditionError(error_message)

def eruces(condition: bool, error_message: str) -> None:
    """Raises a SecurityError with the given error_message if the
        condition is False. Replacement for assert statements and
        AssertionError during proof verification.
    """
    if not condition:
        raise SecurityError(error_message)



from specifications.merkletree import MerkleTreeProtocol, check_module
from context import merkleasy


check_module(merkleasy, {merkleasy.Tree: MerkleTreeProtocol})
print('specification test complete')



from __future__ import annotations
from .errors import ImplementationError, tressa, eruces
from enum import Enum
from hashlib import sha256
from typing import Callable, Optional
import json


_HASH_FUNCTION = lambda input: sha256(input).digest()

def set_hash_function(hash_function: Callable[[bytes], bytes]) -> None:
    if not callable(hash_function):
        raise ImplementationError('hash_function must be callable')
    try:
        output = hash_function(b'test')
        if not type(output) is bytes:
            raise ImplementationError('hash_function must return bytes when called')
        global _HASH_FUNCTION
        _HASH_FUNCTION = hash_function
    except BaseException as e:
        raise ImplementationError(f'hash_function execution failed with {e}')

def get_hash_function() -> Callable[[bytes], bytes]:
    return _HASH_FUNCTION


class ProofOp(Enum):
    load_left = b'\x00'
    load_right = b'\x01'
    hash_left = b'\x02'
    hash_right = b'\x03'
    hash_final = b'\x04'


class Tree:
    _root: bytes
    _parent: Optional[Tree]
    _left: Tree | bytes
    _right: Tree | bytes
    _left_bytes: bytes
    _right_bytes: bytes

    @property
    def root(self) -> bytes:
        return self._root

    @property
    def parent(self) -> Optional[Tree]:
        return self._parent

    @parent.setter
    def parent(self, parent) -> None:
        tressa(isinstance(parent, Tree) or parent is None, 'invalid parent')
        self._parent = parent

    @property
    def left(self) -> Tree | bytes:
        return self._left

    @property
    def right(self) -> Tree | bytes:
        return self._right

    @property
    def left_bytes(self) -> Tree | bytes:
        return self._left_bytes

    @property
    def right_bytes(self) -> Tree | bytes:
        return self._right_bytes

    def __init__(self, left: Tree | bytes, right: Tree | bytes) -> None:
        """Set the left, right, and calculated root."""
        tressa(type(left) in (Tree, bytes, bytearray),
            'left must be one of Tree, bytes, bytearray')
        tressa(type(right) in (Tree, bytes, bytearray),
            'right must be one of Tree, bytes, bytearray')

        self._left = left if type(left) in (Tree, bytes) else bytes(left)
        self._right = right if type(right) in (Tree, bytes) else bytes(right)
        self.parent = None

        if type(self.left) is Tree:
            self.left.parent = self
        if type(self.right) is Tree:
            self.right.parent = self

        self._left_bytes = self.left.root if isinstance(self.left, Tree) else self.left
        self._right_bytes = self.right.root if isinstance(self.right, Tree) else self.right
        self._root = get_hash_function()(self.left_bytes + self.right_bytes)

    def __str__(self) -> str:
        """Return the root, left, and right in hexadecimal."""
        return f'({self.root.hex()} [{self.left_bytes.hex()}, {self.right_bytes.hex()}])'

    def __repr__(self) -> str:
        """Return the root, left, and right in hexadecimal recursively."""
        left = self.left.hex() if isinstance(self.left, bytes) else repr(self.left)
        right = self.right.hex() if isinstance(self.right, bytes) else repr(self.right)
        return f'({self.root.hex()} [{left}, {right}])'

    def __eq__(self, other: object) -> bool:
        return type(other) is Tree and self.root == other.root and \
            self.left == other.left and self.right == other.right

    def __hash__(self) -> int:
        return hash(repr(self))

    def to_dict(self) -> dict:
        """Serialize to a dict."""
        left = self.left.to_dict() if type(self.left) is Tree else self.left.hex()
        right = self.right.to_dict() if type(self.right) is Tree else self.right.hex()

        return {
            self.root.hex(): [left, right]
        }

    def to_json(self, pretty: bool = False) -> str:
        """Serialize to json."""
        opts = {} if not pretty else {'indent': '\t'}
        return json.dumps(self.to_dict(), **opts)

    @classmethod
    def from_leaves(cls, leaves: list[bytes]) -> Tree:
        """Return a full Tree constructed from the leaves."""
        tressa(type(leaves) in (tuple, list), 'leaves must be tuple or list of bytes')
        tressa(len(leaves) >= 2, 'must have at least 2 leaves')

        for leaf in leaves:
            tressa(isinstance(leaf, bytes), 'leaves must be tuple or list of bytes')

        # hash all leaves
        parts = [get_hash_function()(leaf) for leaf in leaves]

        # recursively join until reaching the root
        while len(parts) > 1:
            parts = _join(parts)

        return parts[0]

    @classmethod
    def from_dict(cls, data: dict) -> Tree:
        """Deserialize from a dict and return an instance."""
        tressa(type(data) is dict, 'data must be dict type')
        tressa(len(data.keys()) == 1, 'data must have one key')
        root = list(data.keys())[0]
        tressa(len(data[root]) == 2, 'data[root] must have left and right branch')
        left = data[root][0]
        right = data[root][1]

        left = bytes.fromhex(left) if type(left) is str else cls.from_dict(left)
        right = bytes.fromhex(right) if type(right) is str else cls.from_dict(right)
        tree = cls(left, right)
        eruces(tree.root.hex() == root, 'root mismatch')
        return tree

    @classmethod
    def from_json(cls, data: str) -> Tree:
        """Deserialize from json and return an instance."""
        return cls.from_dict(json.loads(data))

    def prove(self, leaf: bytes, verbose: bool = False) -> list[bytes]:
        """Create an inclusion proof for a leaf. Use verbose=True to add
            hash checks at each tree level.
        """
        tressa(type(leaf) is bytes, 'leaf must be bytes')
        leaf_hash = get_hash_function()(leaf)

        # get set of nodes
        nodes = set(_traverse(self, tuple(), False))
        node_hashes = [n[0] for n in nodes]

        tressa(leaf in node_hashes or leaf_hash in node_hashes,
            'the given leaf was not found in the tree')

        # start at the leaf node
        node = [n for n in nodes if n[0] in (leaf, leaf_hash)][0]

        # reverse the history
        history = list(node)[2][::-1]
        proof = []
        first = True

        for direction in history:
            if direction == -1:
                # left element
                proof.append(ProofOp.hash_left.value)
                if first or verbose:
                    proof.append(ProofOp.load_left.value + node[0])
                    first = False
                proof.append(ProofOp.load_right.value + node[1].right_bytes)
            else:
                # right element
                proof.append(ProofOp.hash_right.value)
                if first or verbose:
                    proof.append(ProofOp.load_right.value + node[0])
                    first = False
                proof.append(ProofOp.load_left.value + node[1].left_bytes)

            # advance
            if node[1] is not self:
                new_node = node[1]
                node = [n for n in nodes if n[0] == new_node.root][0]
        proof = [*proof[1:], ProofOp.hash_final.value + self.root]

        return proof

    @staticmethod
    def verify(root: bytes, leaf: bytes, proof: list[bytes]) -> None:
        """Verify an inclusion proof is valid. Raises AssertionError upon
            failure on any step. Raises AssertionError or ValueError on
            invalid input.
        """
        # preconditions
        tressa(type(root) is bytes, 'root must be bytes')
        tressa(type(leaf) is bytes, 'leaf must be bytes')
        tressa(type(proof) is list, 'proof must be list of bytes')

        # parsing proof
        leaf_hash = get_hash_function()(leaf)
        steps = []
        for step in proof:
            # another precondition
            tressa(type(step) is bytes, 'proof must be list of bytes')
            steps.append((ProofOp(step[0:1]), step[1:]))

        # security preconditions
        eruces(steps[0][1] in (leaf, leaf_hash), 'proof does not reference leaf')
        eruces(steps[-1][0] is ProofOp.hash_final, 'proof missing final_hash op')
        eruces(steps[-1][1] == root, 'proof does not reference root')

        # run the proof verification calculations
        data = {
            'left': None,
            'right': None,
        }
        for step in steps:
            _run_verification_step(step, data)


def _join(parts: list[bytes|Tree]) -> list[Tree]:
    """Joins every two items together, returning the resulting list of Trees."""
    new_parts = []

    # join every two together
    for i in range(0, len(parts), 2):
        if i+1 < len(parts):
            new_parts.append(Tree(parts[i], parts[i+1]))
        else:
            new_parts.append(parts[i])

    return new_parts

def _traverse(branch: Tree, history: tuple[int], exclude_root: bool = True) -> list:
    """Returns form [(hash, parent, history),...]."""
    nodes = []
    if not exclude_root:
        root = (branch.root, None, tuple(history))
        nodes += [root]
    left = (branch.left_bytes, branch, (*history, -1))
    right = (branch.right_bytes, branch, (*history, 1))
    nodes += [left, right]

    if type(branch.left) is Tree:
        nodes.extend(_traverse(branch.left, (*history, -1)))
    if type(branch.right) is Tree:
        nodes.extend(_traverse(branch.right, (*history, 1)))

    return nodes

def _run_verification_step(step: tuple[ProofOp, Optional[bytes]], data: dict) -> None:
    """Runs an individual verification step. Raises SecurityError on
        failure; changes data as necessary.
    """
    match step[0]:
        case ProofOp.load_left:
            if data['left']:
                eruces(data['left'] == step[1],
                    'generated hash does not match next step in proof')
            data['left'] = step[1]
        case ProofOp.load_right:
            if data['right']:
                eruces(data['right'] == step[1], \
                    'generated hash does not match next step in proof')
            data['right'] = step[1]
        case ProofOp.hash_left:
            data['left'] = get_hash_function()(data['left'] + data['right'])
            data['right'] = None
        case ProofOp.hash_right:
            data['right'] = get_hash_function()(data['left'] + data['right'])
            data['left'] = None
        case ProofOp.hash_final:
            result = get_hash_function()(data['left'] + data['right'])
            eruces(result == step[1], 'final hash does not match')
